/* gobby - A GTKmm driven libobby client
 * Copyright (C) 2005 0x539 dev group
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gtkmm/stock.h>
#include <gtkmm/toggleaction.h>
#include <gtkmm/radioaction.h>
#include <obby/buffer.hpp>

#include "features.hpp"
#include "common.hpp"
#include "header.hpp"

namespace {
	Glib::ustring ui_desc = 
		"<ui>"
		"  <menubar name=\"MenuMainBar\">"
		"    <menu action=\"MenuApp\">"
		"      <menuitem action=\"CreateSession\" />"
		"      <menuitem action=\"JoinSession\" />"
		"      <menuitem action=\"QuitSession\" />"
		"      <separator />"
		"      <menuitem action=\"Quit\" />"
		"    </menu>"
		"    <menu action=\"MenuSession\">"
		"      <menuitem action=\"CreateDocument\" />"
		"      <menuitem action=\"OpenDocument\" />"
		"      <menuitem action=\"SaveDocument\" />"
		"      <menuitem action=\"CloseDocument\" />"
		"    </menu>"
#ifdef WITH_GTKSOURCEVIEW
		"    <menu action=\"MenuView\">"
		"      <menuitem action=\"ViewLineNumbers\" />"
		"      <separator />"
		"      <menuitem action=\"ViewLanguageNone\" />"
		"    </menu>"
#endif
		"    <menu action=\"MenuHelp\">"
		"      <menuitem action=\"About\" />"
		"    </menu>"
		"  </menubar>"
		"  <toolbar name=\"ToolMainBar\">"
		"    <toolitem action=\"CreateSession\" />"
		"    <toolitem action=\"JoinSession\" />"
		"    <toolitem action=\"QuitSession\" />"
		"    <separator />"
		"    <toolitem action=\"CreateDocument\" />"
		"    <toolitem action=\"OpenDocument\" />"
		"    <toolitem action=\"SaveDocument\" />"
		"    <toolitem action=\"CloseDocument\" />"
		"  </toolbar>"
		"</ui>";

	void replace_string(Glib::ustring& string,
	                    const Glib::ustring& find,
	                    const Glib::ustring& replace)
	{
		Glib::ustring::size_type pos = 0;
		while((pos = string.find(find, pos)) != Glib::ustring::npos)
		{
			string.replace(pos, find.length(), replace);
			pos += replace.length();
		}
	}
	
	void remove_entities(Glib::ustring& string)
	{
		// Note that this file needs to be UTF-8 encoded!
		replace_string(string, "ä", "auml");
		replace_string(string, "Ä", "Auml");
		replace_string(string, "ö", "ouml");
		replace_string(string, "Ö", "Ouml");
		replace_string(string, "ü", "uuml");
		replace_string(string, "Ü", "Uuml");
		replace_string(string, "ß", "szlig");
		replace_string(string, "<", "lt");
		replace_string(string, ">", "gt");
		replace_string(string, "\"", "quot");

		// TODO: Remove some more dangerous characters
	}
}

Gobby::Header::Error::Error(Code error_code, const Glib::ustring& error_message)
 : Glib::Error(g_quark_from_static_string("GOBBY_HEADER_ERROR"),
               static_cast<int>(error_code), error_message)
{
}

Gobby::Header::Error::Code Gobby::Header::Error::code() const
{
	return static_cast<Code>(gobject_->code);
}

Gobby::Header::Header(const Folder& folder)
 : m_ui_manager(Gtk::UIManager::create() ),
   m_group_app(Gtk::ActionGroup::create() )
#ifdef WITH_GTKSOURCEVIEW
   , m_toggle_language(false), m_toggle_line_numbers(false)
#endif
{
	// Add basic menu
	m_ui_manager->add_ui_from_string(ui_desc);

	// App menu
	m_group_app->add(Gtk::Action::create("MenuApp", _("Gobby")) );

	// Create session
	m_group_app->add(
		Gtk::Action::create(
			"CreateSession",
			Gtk::Stock::NETWORK,
			_("Create session"),
			_("Opens a new obby session")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_session_create
		)
	);

	// Join session
	m_group_app->add(
		Gtk::Action::create(
			"JoinSession",
			Gtk::Stock::CONNECT,
			_("Join session"),
			_("Join an existing obby session")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_session_join
		)
	);

	// Quit session
	m_group_app->add(
		Gtk::Action::create(
			"QuitSession",
			Gtk::Stock::DISCONNECT,
			_("Quit session"),
			_("Leaves the currently running obby session")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_session_quit
		)
	);

	// Quit application
	m_group_app->add(
		Gtk::Action::create(
			"Quit",
			Gtk::Stock::QUIT,
			_("Quit"),
			_("Quits the application")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_quit
		)
	);

	// Session menu
	m_group_app->add(Gtk::Action::create("MenuSession", _("Session")) );

	// Create document
	m_group_app->add(
		Gtk::Action::create(
			"CreateDocument",
			Gtk::Stock::NEW,
			_("Create document"),
			_("Creates a new document")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_document_create
		)
	);

	// Open document
	m_group_app->add(
		Gtk::Action::create(
			"OpenDocument",
			Gtk::Stock::OPEN,
			_("Open document"),
			_("Loads a file into a new document")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_document_open
		)
	);

	// Save document
	m_group_app->add(
		Gtk::Action::create(
			"SaveDocument",
			Gtk::Stock::SAVE,
			_("Save document"),
			_("Saves a document into a file")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_document_save
		)
	);

	// Close document
	m_group_app->add(
		Gtk::Action::create(
			"CloseDocument",
			Gtk::Stock::CLOSE,
			_("Close document"),
			_("Closes an open document")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_document_close
		)
	);

#ifdef WITH_GTKSOURCEVIEW
	// Documents menu
	m_group_app->add(Gtk::Action::create("MenuView", _("View")) );

	// Show line numbers
	m_group_app->add(
		Gtk::ToggleAction::create(
			"ViewLineNumbers",
			_("Line numbers"),
			_("Whether to show line numbers for this document")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_document_line_numbers
		)
	);

	// A kind of hack to ensure that
	// Gtk::SourceLanguage::sourcelanguage_class_.init() is called.
	// See the TODO item in Glib::wrap(GtkSourceLanguage*, bool) in
	// sourcelanguage.cpp
	GtkSourceLanguage* lang = NULL;
	Glib::wrap(lang, false);

	// Get languages manager
	Glib::RefPtr<Gtk::SourceLanguagesManager> lang_manager =
		folder.get_lang_manager();

	// Get available languages
	std::list<Glib::RefPtr<Gtk::SourceLanguage> > lang_list =
		lang_manager->get_available_languages();
	
	// Sort languages by name
	lang_list.sort(&Header::language_sort_callback);

	// Add None-Language
	m_group_app->add(
		Gtk::RadioAction::create(
			m_lang_group,
			"ViewLanguageNone",
			"None",
			"Unselects the current language"
		),
		sigc::bind(
			sigc::mem_fun(
				*this,
				&Header::on_app_document_language
			),
			Glib::RefPtr<Gtk::SourceLanguage>()
		)
	);

	// Add languages
	std::list<Glib::RefPtr<Gtk::SourceLanguage> >::const_iterator iter;
	for(iter = lang_list.begin(); iter != lang_list.end(); ++ iter)
	{
		// Get current language 
		Glib::RefPtr<Gtk::SourceLanguage> language = *iter;
		Glib::ustring language_xml_name = language->get_name();

		// Add language to action group
		remove_entities(language_xml_name);
		m_group_app->add(
			Gtk::RadioAction::create(
				m_lang_group,
				"ViewLanguage" + language_xml_name, 
				language->get_name(),
				_("Selects ") + language->get_name() +
					_(" as language")
			),
			sigc::bind(
				sigc::mem_fun(
					*this,
					&Header::on_app_document_language
				),
				language
			)
		);

		// Add menu item to UI
		Glib::ustring xml_desc =
			"<ui>"
			"  <menubar name=\"MenuMainBar\">"
			"    <menu action=\"MenuView\">"
			"	<menuitem action=\"ViewLanguage"
				+ language_xml_name + "\" />"
			"    </menu>"
			"  </menubar>"
			"</ui>";

		m_ui_manager->add_ui_from_string(xml_desc);
	}
#endif

	// Help menu
	m_group_app->add(Gtk::Action::create("MenuHelp", _("Help")) );

	// Display about dialog
	m_group_app->add(
		Gtk::Action::create(
			"About",
			Gtk::Stock::ABOUT,
			_("About"),
			_("Shows Gobby's copyright and credits")
		),
		sigc::mem_fun(
			*this,
			&Header::on_app_about
		)
	);

	m_ui_manager->insert_action_group(m_group_app);

	m_menubar = static_cast<Gtk::MenuBar*>(
		m_ui_manager->get_widget("/MenuMainBar") );
	m_toolbar = static_cast<Gtk::Toolbar*>(
		m_ui_manager->get_widget("/ToolMainBar") );

	if(m_menubar == NULL)
		throw Error(Error::MENUBAR_MISSING,
			"XML UI definition lacks menubar" );
	if(m_toolbar == NULL)
		throw Error(Error::TOOLBAR_MISSING,
			"XML UI definition lacks toolbar" );

	pack_start(*m_menubar, Gtk::PACK_SHRINK);
	pack_start(*m_toolbar, Gtk::PACK_SHRINK);

	// Without having joined a session, there is no support for documents
	// or the possibility to quit it.
	m_group_app->get_action("CreateDocument")->set_sensitive(false);
	m_group_app->get_action("OpenDocument")->set_sensitive(false);
	m_group_app->get_action("SaveDocument")->set_sensitive(false);
	m_group_app->get_action("CloseDocument")->set_sensitive(false);
	m_group_app->get_action("QuitSession")->set_sensitive(false);
#ifdef WITH_GTKSOURCEVIEW
	m_group_app->get_action("MenuView")->set_sensitive(false);
#endif

	// Connect to folder's signals
	folder.tab_switched_event().connect(
		sigc::mem_fun(*this, &Header::on_folder_tab_switched) );
}

Gobby::Header::~Header()
{
}

Glib::RefPtr<Gtk::AccelGroup> Gobby::Header::get_accel_group()
{
	return m_ui_manager->get_accel_group();
}

Glib::RefPtr<const Gtk::AccelGroup> Gobby::Header::get_accel_group() const
{
	return m_ui_manager->get_accel_group();
}

void Gobby::Header::disable_document_actions()
{
	m_group_app->get_action("SaveDocument")->set_sensitive(false);
	m_group_app->get_action("CloseDocument")->set_sensitive(false);
#ifdef WITH_GTKSOURCEVIEW
	m_group_app->get_action("MenuView")->set_sensitive(false);
#endif
}

Gobby::Header::signal_session_create_type
Gobby::Header::session_create_event() const
{
	return m_signal_session_create;
}

Gobby::Header::signal_session_join_type
Gobby::Header::session_join_event() const
{
	return m_signal_session_join;
}

Gobby::Header::signal_session_quit_type
Gobby::Header::session_quit_event() const
{
	return m_signal_session_quit;
}

Gobby::Header::signal_document_create_type
Gobby::Header::document_create_event() const
{
	return m_signal_document_create;
}

Gobby::Header::signal_document_open_type
Gobby::Header::document_open_event() const
{
	return m_signal_document_open;
}

Gobby::Header::signal_document_save_type
Gobby::Header::document_save_event() const
{
	return m_signal_document_save;
}

Gobby::Header::signal_document_close_type
Gobby::Header::document_close_event() const
{
	return m_signal_document_close;
}

#ifdef WITH_GTKSOURCEVIEW
Gobby::Header::signal_document_line_numbers_type
Gobby::Header::document_line_numbers_event() const
{
	return m_signal_document_line_numbers;
}

Gobby::Header::signal_document_language_type
Gobby::Header::document_language_event() const
{
	return m_signal_document_language;
}
#endif

Gobby::Header::signal_about_type
Gobby::Header::about_event() const
{
	return m_signal_about;
}

Gobby::Header::signal_quit_type
Gobby::Header::quit_event() const
{
	return m_signal_quit;
}

void Gobby::Header::obby_start()
{
	// Begin of obby session: Disable create/join buttons, enable quit
	m_group_app->get_action("CreateSession")->set_sensitive(false);
	m_group_app->get_action("JoinSession")->set_sensitive(false);
	m_group_app->get_action("QuitSession")->set_sensitive(true);

	// Enable document buttons
	m_group_app->get_action("CreateDocument")->set_sensitive(true);
	m_group_app->get_action("OpenDocument")->set_sensitive(true);

	// Document actions will be activated from the insert_document event
	m_group_app->get_action("SaveDocument")->set_sensitive(false);
	m_group_app->get_action("CloseDocument")->set_sensitive(false);
#ifdef WITH_GTKSOURCEVIEW
	m_group_app->get_action("MenuView")->set_sensitive(false);
#endif
}

void Gobby::Header::obby_end()
{
	// End of obby session: Enable create/join buttons, disable quit
	m_group_app->get_action("CreateSession")->set_sensitive(true);
	m_group_app->get_action("JoinSession")->set_sensitive(true);
	m_group_app->get_action("QuitSession")->set_sensitive(false);

	// Disable document buttons
	m_group_app->get_action("CreateDocument")->set_sensitive(false);
	m_group_app->get_action("OpenDocument")->set_sensitive(false);

	// Leave document actions like SaveDocument and CloseDocument as they
	// were to allow the user to save documents.
}

void Gobby::Header::obby_user_join(obby::user& user)
{
}

void Gobby::Header::obby_user_part(obby::user& user)
{
}

void Gobby::Header::obby_document_insert(obby::document& document)
{
	// Now we have at least one document open, so we could activate the
	// document actions.
	m_group_app->get_action("SaveDocument")->set_sensitive(true);
	m_group_app->get_action("CloseDocument")->set_sensitive(true);
#ifdef WITH_GTKSOURCEVIEW
	m_group_app->get_action("MenuView")->set_sensitive(true);
#endif
}

void Gobby::Header::obby_document_remove(obby::document& document)
{
	if(document.get_buffer().document_count() == 1)
	{
		// The document which is currently removed is the only
		// existing document? Disable document actions then.
		m_group_app->get_action("SaveDocument")->set_sensitive(false);
		m_group_app->get_action("CloseDocument")->set_sensitive(false);
#ifdef WITH_GTKSOURCEVIEW
		m_group_app->get_action("MenuView")->set_sensitive(false);
#endif
	}
}

void Gobby::Header::on_app_session_create()
{
	m_signal_session_create.emit();
}

void Gobby::Header::on_app_session_join()
{
	m_signal_session_join.emit();
}

void Gobby::Header::on_app_session_quit()
{
	m_signal_session_quit.emit();
}

void Gobby::Header::on_app_document_create()
{
	m_signal_document_create.emit();
}

void Gobby::Header::on_app_document_open()
{
	m_signal_document_open.emit();
}

void Gobby::Header::on_app_document_save()
{
	m_signal_document_save.emit();
}

void Gobby::Header::on_app_document_close()
{
	m_signal_document_close.emit();
}

#ifdef WITH_GTKSOURCEVIEW
void Gobby::Header::on_app_document_line_numbers()
{
	// Are we toggling line numbers manually through a tab change or
	// something? Then we do not have to emit this signal because we
	// read the new status from the document and need not to set the
	// document's line number status to the value it currently has ;)
	if(!m_toggle_line_numbers)
		m_signal_document_line_numbers.emit();
}

void Gobby::Header::on_app_document_language(
	Glib::RefPtr<Gtk::SourceLanguage> lang
)
{
	// Same as above
	if(!m_toggle_language)
		m_signal_document_language.emit(lang);
}
#endif

void Gobby::Header::on_app_about()
{
	m_signal_about.emit();
}

void Gobby::Header::on_app_quit()
{
	m_signal_quit.emit();
}

void Gobby::Header::on_folder_tab_switched(Document& document)
{
#ifdef WITH_GTKSOURCEVIEW
	// We are toggling line numbers
	m_toggle_line_numbers = true;
	m_toggle_language = true;

	// Set current line number state
	Glib::RefPtr<Gtk::ToggleAction>::cast_static<Gtk::Action>(
		m_group_app->get_action("ViewLineNumbers")
	)->set_active(document.get_show_line_numbers() );

	// Set current language
	Glib::ustring langname = document.get_language() ?
		document.get_language()->get_name() : "None";
	remove_entities(langname);
	Glib::RefPtr<Gtk::RadioAction>::cast_static<Gtk::Action>(
		m_group_app->get_action("ViewLanguage" + langname)
	)->set_active();

	// We are no more toggling
	m_toggle_line_numbers = false;
	m_toggle_language = false;
#endif
}

#ifdef WITH_GTKSOURCEVIEW
bool Gobby::Header::language_sort_callback(
	const Glib::RefPtr<Gtk::SourceLanguage>& lang1,
	const Glib::RefPtr<Gtk::SourceLanguage>& lang2
)
{
	return lang1->get_name() < lang2->get_name();
}
#endif
